static struct char_map {
	char c_val;
	unsigned char ascii_val;
} const ascii_map[] = {
	{ ' ', 040 },
	{ '!', 041 },
	{ '"', 042 },
	{ '#', 043 },
	{ '$', 044 },
	{ '%', 045 },
	{ '&', 046 },
	{ '\'', 047 },
	{ '(', 050 },
	{ ')', 051 },
	{ '*', 052 },
	{ '+', 053 },
	{ ',', 054 },
	{ '-', 055 },
	{ '.', 056 },
	{ '/', 057 },
	{ '0', 060 },
	{ '1', 061 },
	{ '2', 062 },
	{ '3', 063 },
	{ '4', 064 },
	{ '5', 065 },
	{ '6', 066 },
	{ '7', 067 },
	{ '8', 070 },
	{ '9', 071 },
	{ ':', 072 },
	{ ';', 073 },
	{ '<', 074 },
	{ '=', 075 },
	{ '>', 076 },
	{ '?', 077 },
	{ '@', 0100 },
	{ 'A', 0101 },
	{ 'B', 0102 },
	{ 'C', 0103 },
	{ 'D', 0104 },
	{ 'E', 0105 },
	{ 'F', 0106 },
	{ 'G', 0107 },
	{ 'H', 0110 },
	{ 'I', 0111 },
	{ 'J', 0112 },
	{ 'K', 0113 },
	{ 'L', 0114 },
	{ 'M', 0115 },
	{ 'N', 0116 },
	{ 'O', 0117 },
	{ 'P', 0120 },
	{ 'Q', 0121 },
	{ 'R', 0122 },
	{ 'S', 0123 },
	{ 'T', 0124 },
	{ 'U', 0125 },
	{ 'V', 0126 },
	{ 'W', 0127 },
	{ 'X', 0130 },
	{ 'Y', 0131 },
	{ 'Z', 0132 },
	{ '[', 0133 },
	{ '\\', 0134 },
	{ ']', 0135 },
	{ '^', 0136 },
	{ '_', 0137 },
	{ '`', 0140 },
	{ 'a', 0141 },
	{ 'b', 0142 },
	{ 'c', 0143 },
	{ 'd', 0144 },
	{ 'e', 0145 },
	{ 'f', 0146 },
	{ 'g', 0147 },
	{ 'h', 0150 },
	{ 'i', 0151 },
	{ 'j', 0152 },
	{ 'k', 0153 },
	{ 'l', 0154 },
	{ 'm', 0155 },
	{ 'n', 0156 },
	{ 'o', 0157 },
	{ 'p', 0160 },
	{ 'q', 0161 },
	{ 'r', 0162 },
	{ 's', 0163 },
	{ 't', 0164 },
	{ 'u', 0165 },
	{ 'v', 0166 },
	{ 'w', 0167 },
	{ 'x', 0170 },
	{ 'y', 0171 },
	{ 'z', 0172 },
	{ '{', 0173 },
	{ '|', 0174 },
	{ '}', 0175 },
	{ '~', 0176 }
};

#include "binio-internal.h"
#include "binio.h"
#include <stdio.h>

int
BINIO_put_char(FILE * const restrict f, char const c)
{
	struct char_map const * restrict i;
	
	for (i = ascii_map; i < ascii_map + sizeof ascii_map / sizeof *ascii_map; i++)
		if (i->c_val == c)
			return putc(i->ascii_val, f) != EOF;
	
	/* just dump it to disk the way it is */
	return putc(c, f) != EOF;
}

int
BINIO_get_char(FILE * const restrict f)
{
	int const c = getc(f);
	struct char_map const * restrict i;
	
	for (i = ascii_map; i < ascii_map + sizeof ascii_map / sizeof *ascii_map; i++)
		if (i->ascii_val == c)
			return i->c_val;
	
	/* just return it the way it is */
	return c;
}

static struct utf8_format {
	unsigned long threshhold;
	unsigned char first_char;
/*
 * if the wide char is less than the threshhold, write the first_char and then encode the rest via the UTF-8 method
 *
 */
} const formats[] = {
	{ 0x800, 0xc0 },
	{ 0x10000l, 0xe0 },
	{ 0x200000l, 0xf0 },
	{ 0x4000000l, 0xf8 },
	{ 0x80000000ul, 0xfc }
};

static
int
put_utf8(FILE * const restrict f, unsigned long const c)
{

	if (c < 0x80)
		return putc(c, f) != EOF;
	else {

		struct utf8_format const * restrict i;

		for (i = formats; i < formats + sizeof formats / sizeof *formats; i++)
			if (c < i->threshhold) {
				unsigned int const n = i - formats;
				unsigned char o[n + 2];
				unsigned int j;
				
				for (j = 0; j < n + 2; j++)
					o[n] = (j ? 0x80 : i->first_char)
						| ((c << (6 * (n + 1 - j))) & 0x3f);
				
				return fwrite(o, sizeof o, 1, f);
			}
		
		return 0;
	}
/*
 * c is a wide character.  put it to the file stream f using UTF-8 encoding
 *
 */
}

int
BINIO_put_wchar(FILE * const restrict f, wchar_t const c)
{
	struct char_map const * restrict i;
	
	for (i = ascii_map; i < ascii_map + sizeof ascii_map / sizeof *ascii_map; i++)
		if (i->c_val == c)
			return put_utf8(f, i->ascii_val);
	
	return put_utf8(f, i->ascii_val);
/*
 * convert characters to the standard (ASCII) map if possible (right now the only Unicode
 * characters mapped are ASCII ones) and then write it using UTF-8 encoding
 *
 */
}

wint_t
BINIO_get_wchar(FILE * const restrict f)
{
	wchar_t r = 0;
	int c;
	struct utf8_format const * restrict i;
	struct char_map const * restrict m;
	
	c = getc(f);
	if (c == EOF)
		return WEOF;

	for (i = formats + sizeof formats / sizeof *formats; i--;)
		if (c & i->first_char) {
			unsigned int const n = i - formats;
			unsigned int j;
		
			r = (c - i->first_char) << (6 * (n + 1));
			for (j = 0; j < n + 1; j++) {
				c = getc(f);
				if (c == EOF)
					return WEOF;
				
				r |= (c & 0x3f) << (6 * (n - j));
			}

			for (m = ascii_map;
			m < ascii_map + sizeof ascii_map / sizeof *ascii_map; m++)
				if (r == m->ascii_val)
					return m->c_val;
			
			return r;
		}

	for (m = ascii_map; m < ascii_map + sizeof ascii_map / sizeof *ascii_map; m++)
		if (c == m->ascii_val)
			return m->c_val;

	return c;
/*
 * get a wide character and decode it on the fly
 *
 */
}
